/*********************************************************************************
* UCI-CHANGEABLE CONSTANTS: these constants are changeable via the UCI           *
*   'setoption' command. This numbers denoted here are only initialisation       *
*   values.                                                                      *
*********************************************************************************/

#pragma once

/* aspiration windows */
static int first_aspiration_window       = 15;
static int second_aspiration_window      = 50;
/* conversion from [kd] (king defects) to [cp] (centipawns). See <init.c> for details. */
static int king_safety_care_1st          = 90;
static int king_safety_care_2nd          = 20;
static int king_safety_upper_limit       = 200;


/*********************************************************************************
* EVALUATION CONSTANTS: That's the heart of the evaluation function. I deter-    *
*   mined these values with steady playing around and also looking at other      *
*   open-source engines and books. You're free to change them and if you think   *
*   you've found any room for improvement, feel free to drop me a mail to        *
*   philemon@cycovery.com - you're welcome!                                      *
*                                                                                *
*   all constants are denoted in centipawns.                                     * 
*********************************************************************************/

/* 1. material evaluation: values of every piece class */
#define V_PAWN    100
#define V_KNIGHT  300
#define V_BISHOP  300
#define V_ROOK    500
#define V_QUEEN   900     


/* 2a. positional evaluation: boni/mali for every kind of positional advantage, independent from
                              the position of the piece that gains the points. */                           
#define ENDGAME_THRESHOLD        (1400)

#define TWIN_BISHOPS               (25) /* according to some chess books rating 
                                         a bishop pair equal to a quarter pawn */                                 
#define ENDGAME_BISHOP_BONUS       (10)

#define FIANCHETTO                 (20)

#define ROOK_OPEN_LINE             (20)

#define ROOK_HALFOPEN_LINE          (5)

#define SINGLE_ROOK_7TH_2ND        (25)

#define ROOK_7TH_2ND_ABSOLUTE      (15)

#define REINFORCED_ROOK_7TH_2ND    (40) /* will be counted twice for 2 rooks -> 80 */

#define REINFORCING_QUEEN_7TH_2ND  (20) /* will be added to the single-rook bonus */

#define MULTI_PAWN                (-10) /* will be counted n times for n pawns (mostly 2) */

#define ISOLANI                   (-20)

#define COVERED_PASSED_7TH         (90) /* covered passed pawns where the more advanced one
                                           is sitting on the 7th rank (e.g. d6,e7) */
#define PASSED_ROOK_SUPPORT        (10)

#define PREMATURE_QUEEN           (-17) /* that's more than the maximum of points the queen can gain
                                           with the positional tables */
                                           
#define CASTLING_LOSS_PENALTY     (-40)

#define UNCASTLED_OPEN_E_LINE     (-20)

#define BISHOP_BASELINE_CAGED     (-12)


/* 2b. positional evaluation: boni/mali for every kind of positional pawn advantage, dependent on
                              the current rank of the pawn that gains the points. */                              
static int passedpawn_bonus[8]={0,2,10,20,40,60,70,0};
static int covered_bonus[8]={0,0,4,6,8,12,16,0};
static int twinpawn_bonus[8]={0,0,1,2,3,4,7,0}; /* will be counted twice */
static int passedpawnwithking_bonus[8]={0,0,0,0,10,50,80,0};
static int blockedpassedpawn_malus[8]={0,0,-8,-16,-32,-45,-58,0};


/* 2c. positional evaluation: penalties for king danger, denoted in kd (king defects) */
#define KD_INTACT_REFUGE           (-5)
#define KD_THREAT_ANTIFIANCHETTO    (5)
#define KD_THREAT_TROJANHORSE      (50)

static int open_line_danger[4]={0,8,12,16};
static int halfopen_line_danger[4]={0,4,6,8};
static int sidepawn_advance_danger[7]={0,1,2,3,4,5,6};


/*********************************************************************************
* INTERFACE-RELATED CONSTANTS: If you consider the current symbols as            *
*   unsuitable, feel free to change them. They're dependant from the OS due to   *
*   lack of UTF-8 support in many consoles.                                      *
*********************************************************************************/
static const char pieceSymbol[7] = { ' ', /* <- array starts at value 0, figure-constants start at value 1 */
  #if defined(CARBON)
    161, /* <- pawn */  88, /* <- knight */ 198, /* <- bishop */
    184, /* <- rook */ 165, /* <- queen  */ 240  /* <- king   */
	#elif defined(BCC32)
		248, /* <- pawn */  88, /* <- knight */  30, /* <- bishop */
		210, /* <- rook */   4, /* <- queen  */   1  /* <- king   */
	#else
		'B', /* <- pawn */ 'S', /* <- knight */ 'L', /* <- bishop */
		'T', /* <- rook */ 'D', /* <- queen  */ 'K'  /* <- king   */
	#endif
};


/*********************************************************************************
* UNCHANGEABLE CONSTANTS: The best way to crash the engine immediately or to     *
*   cause massive problems is to change one of the following constants. Support  *
*   not included. :P                                                             *
*********************************************************************************/
#define MAXMOVES                100
#define MAXPLY                  200
#define MAXPASSEDSITUATIONS     500

#define ITERATE_NONE           0
#define ITERATE_UNLIMITED      1
#define ITERATE_DEPTH_LIMITED  2
#define ITERATE_TIME_LIMITED   3
#define ITERATE_NODE_LIMITED   4
#define ITERATE_TIME_PRESSURE  5

#define BUFSIZE   4096
#define MAXINT 		32767

#define MATE_THRESHOLD      30000

#define HASH_UNKNOWN        0
#define HASH_UPPERBOUND     1
#define HASH_LOWERBOUND     2
#define HASH_EXACT          3

#define RESULT_OPEN               0
#define RESULT_MATED_WHITE        1
#define RESULT_MATED_BLACK        2
#define RESULT_DRAW_STALE        10
#define RESULT_DRAW_MATERIAL     11
#define RESULT_DRAW_REPETITION   12
#define RESULT_DRAW_50MOVE       13

#define INPUT_QUERY_FREQUENCY    10000

#define OPENING      -1
#define MIDDLEGAME    0
#define ENDGAME       1

#define QUEENSIDE		0
#define KINGSIDE		1

static const int piece_value[7] = {0, V_PAWN, V_KNIGHT, V_BISHOP, V_ROOK, V_QUEEN, MAXINT};

enum { PHRI, UCI } interfaces;
enum { NORMAL_MOVE, PROMOTION, DOUBLE_STEP, ENPASSANT_CAPTURE, CASTLING, KING_CAPTURE } move_types;
enum { PAWN=1, KNIGHT=2, BISHOP=3, ROOK=4, QUEEN=5, KING=6 } figures;
enum { A=0, B=1, C=2, D=3, E=4, F=5, G=6, H=7} lines;
enum { P_BLACK=0, P_WHITE=1, P_BOTH=2 } colors;
enum { A1, B1, C1, D1, E1, F1, G1, H1,
       A2, B2, C2, D2, E2, F2, G2, H2,
       A3, B3, C3, D3, E3, F3, G3, H3,
       A4, B4, C4, D4, E4, F4, G4, H4,
       A5, B5, C5, D5, E5, F5, G5, H5,
       A6, B6, C6, D6, E6, F6, G6, H6,
       A7, B7, C7, D7, E7, F7, G7, H7,
       A8, B8, C8, D8, E8, F8, G8, H8 } squares;