
#include "const.h"
#include "types.h"
#include "globals.h"
#include "macros.h"
#include "proto.h"
#include "matrices.h"


/*********************************************************************************
* DOMOVE: This function executes any given move and stores necessary facts in    *
*   the backup structure *backup to be able to reconstruct the original board    *
*   situation with UndoMove().                                                   *
*********************************************************************************/
void DoMove(GAME *g, MOVE move, SITBACKUP *backup) {
	int src_figure       = (move)&7;              
	int src              = (move>>=3)&63;         
	int dst              = (move>>=6)&63;
	int move_type        = (move>>=6)&7;
	int promotion_figure = (move>>=3)&3;
	int captured         = (move>>=2)&7;
	int dst_figure;

	backup->src_figure=src_figure;
	backup->src=src;
	backup->dst=dst;
	backup->move_type=move_type;
	backup->captured=captured;
	backup->enpassant=g->sit.enpassant;
	backup->rule50counter=g->sit.rule50counter;
	backup->last_irreversible_move=g->sit.last_irreversible_move;

	g->last_destination[g->ply]=dst;
	
	if(captured || src_figure==PAWN) {
		g->sit.last_irreversible_move=g->halfmove_index+g->ply;
		g->sit.rule50counter=1;
	} else {
		g->sit.rule50counter++;
	}
	
	/* clear en passant square */
	if(g->sit.enpassant) {
		XORZobristEP(g->sit.enpassant);
		g->sit.enpassant=0;
	}

	/* check for special pawn moves */
	if(move_type==DOUBLE_STEP) {
		g->sit.enpassant=LSB(pawn_nonattacks_singlestep[src][Color]);
		XORZobristEP(g->sit.enpassant);
	}
	if(move_type==ENPASSANT_CAPTURE) {
		XOREnemySituationBit(PAWN,pawns,dst+ColCmp(-8,8));
		Piece(dst+ColCmp(-8,8))=0;
		g->sit.material_score[!Color]-=V_PAWN;
		g->sit.pawn_count[!Color]-=1;
	}

	/* check for castling */
	XORZobristCastling; /* first time: erase the castling key */
	backup->castling_disallowed[Color]=g->sit.castling_disallowed[Color];
	if(move_type==CASTLING) {
		g->sit.castling_disallowed[Color]|=3;
		if(Line(dst)==6) {
			XOROwnSituationBit(ROOK,rooks,ColCmp(H1,H8));
			XOROwnSituationBit(ROOK,rooks,ColCmp(F1,F8));
			Piece(ColCmp(H1,H8))=0;
			Piece(ColCmp(F1,F8))=ColCmp(ROOK,-ROOK);
		} else if(Line(dst)==2) {
			XOROwnSituationBit(ROOK,rooks,ColCmp(A1,A8));
			XOROwnSituationBit(ROOK,rooks,ColCmp(D1,D8));
			Piece(ColCmp(A1,A8))=0;
			Piece(ColCmp(D1,D8))=ColCmp(ROOK,-ROOK);
		} else
			Error(errDetails,"Illegal Castling, Line(dst) is neither 6 nor 2!");
	}
		
	/* clear old piece */
	switch(abs(Piece(src))) {
		case PAWN:   XOROwnSituationBit(PAWN,pawns,src);     break;
		case KNIGHT: XOROwnSituationBit(KNIGHT,knights,src); break;
		case BISHOP: XOROwnSituationBit(BISHOP,bishops,src); break;
		case ROOK:   XOROwnSituationBit(ROOK,rooks,src);     break;
		case QUEEN:  XOROwnSituationBit(QUEEN,queens,src);   break;
		case KING:   XOROwnFigureset(src);                   
		             XORZobristOwnKing(src);                 break;
	}
	Piece(src)=0;

	/* remove captured figure */
	if(captured) {
		switch(captured) {
			case PAWN:   XOREnemySituationBit(PAWN,pawns,dst);     break;
			case KNIGHT: XOREnemySituationBit(KNIGHT,knights,dst); break;
			case BISHOP: XOREnemySituationBit(BISHOP,bishops,dst); break;
			case ROOK:   XOREnemySituationBit(ROOK,rooks,dst);     break;
			case QUEEN:  XOREnemySituationBit(QUEEN,queens,dst);   break;
			case KING:   Error(errDetails,"king capture detected!");          break;
		}
		g->sit.material_score[!Color]-=piece_value[captured];
		if(captured==PAWN)
			g->sit.pawn_count[!Color]-=1;
	}
	
	/* determine & set the new figure */
	if(move_type==PROMOTION) {
		dst_figure=promotion_figure+2; /* scale from 2-bit (0..3) to correct piece numbers (2..5) */
		g->sit.material_score[Color]+=piece_value[abs(dst_figure)]-V_PAWN;
		g->sit.pawn_count[Color]-=1;
	}
	else
		dst_figure=src_figure;
	backup->dst_figure=dst_figure;
	
	switch(dst_figure) {
		case PAWN:   XOROwnSituationBit(PAWN,pawns,dst);     break;
		case KNIGHT: XOROwnSituationBit(KNIGHT,knights,dst); break;
		case BISHOP: XOROwnSituationBit(BISHOP,bishops,dst); break;
		case ROOK:   XOROwnSituationBit(ROOK,rooks,dst);     break;
		case QUEEN:  XOROwnSituationBit(QUEEN,queens,dst);   break;
		case KING:   XOROwnFigureset(dst);                   
		             XORZobristOwnKing(dst);                 break;
	}
	Piece(dst)=dst_figure*(ColCmp(1,-1));

	/* update castling permissions */
	if (src_figure==ROOK && src==ColCmp(H1,H8))
		g->sit.castling_disallowed[Color]|=1;
	
	if (src_figure==ROOK && src==ColCmp(A1,A8))
		g->sit.castling_disallowed[Color]|=2;
		
	/* set new king's position */
	if (src_figure==KING) {
		g->sit.king[Color]=dst;
		g->sit.castling_disallowed[Color]|=3;
	}
	/* second time: set the castling key */
	XORZobristCastling;
	
	/* put the current move on the stack */
	UpdateHalfmoveIndex;
	g->passed_zobrist[g->halfmove_index+g->ply-1]=g->sit.zobrist;
}