
#include "const.h"
#include "types.h"
#include "globals.h"
#include "macros.h"
#include "proto.h"
#include "matrices.h"


/*********************************************************************************
* EVALUATION ROUTINE: That's the backbone of the whole program. It's mainly      *
*   based on A.Bartashnikovs training CDs and Aaron Nimzowitsch's famous book    *
*   "my system". Despite a lot of testing and playing aroung, I know it's still  *
*   far away from being good. So if you think you've found any room for          *
*   improvement, please let me know and drop me a mail to philemon@cycovery.com, *
*   just now!                                                                    *
*                                                                                *
*   See "const.h" for all the constants being used here.                         * 
*********************************************************************************/
int Evaluate(GAME *g, int print_score) {
	int pos_score[2]={0,0};
	int col,i,score;
	int line;
	BITBOARD remaining_figures;
	BITBOARD tmp_bb;
	BITBOARD pos_mask;
	char pos;
	int num_pieces;
	int color_save=Color;
	/* pawn variables */
	int pawns_on_line[2][8];   /* needed for half/open lines and some pawn evaluations */
	unsigned char passedpawn_lines[2]={0,0};
	int pos_score_table;
	int open_lines[2]={0,0};
	int halfopen_lines[2][2]={{0,0},{0,0}};
	int advanced_sidepawns[2][2]={{0,0},{0,0}};
	int side_danger[2][2]={{0,0},{0,0}};
	int defective_refuge[2][2]={{0,0},{0,0}};
	int multipawn_penalty[2]={0,0};
	
	/* determine the game phase */
	g->phase=DeterminePhase(g);
	
	/* start with the pure material balance */
	EvHeader("MATERIAL");
	if(print_score) {
		PHRIStaticEvaluationRow("total white material",-1,P_WHITE,g->sit.material_score[P_WHITE]);
		PHRIStaticEvaluationRow("total black material",-1,P_BLACK,g->sit.material_score[P_BLACK]);
	}
	score=g->sit.material_score[P_WHITE]-g->sit.material_score[P_BLACK];

	/* collect information about lines, advanced pawns, multipawns, etc. */
	for (line=A;line<=H;line++)
		if (m_line[line]&(g->sit.pawns[P_BLACK]|g->sit.pawns[P_WHITE])) {
			/* no open line */
			for (col=P_WHITE; col>=P_BLACK; col--) {
				Color=col;
				if (tmp_bb=m_line[line]&g->sit.pawns[col]) {
					/* no half-open line */
					if ((pawns_on_line[col][line]=CntBits(tmp_bb))>1) {
						multipawn_penalty[col]+=pawns_on_line[col][line]*MULTI_PAWN;
					}
					if (!EvCenterLine)
						/* A,B,C or F,G,H line */
						if (!(mask64[SqVal(line,EvColCmp(1,6))]&OwnPawns)) {
							if (!(mask64[SqVal(line,EvColCmp(2,5))]&OwnPawns))
								/* no pawn on 2nd & 3rd rank */	
								advanced_sidepawns[col][EvSide]+=2;
							else
								/* no pawn on 2nd rank */
								advanced_sidepawns[col][EvSide]+=1;
							EvDisruptRefuge;
						}
				} else {
					if (!EvCenterLine)
						halfopen_lines[col][EvSide]++;
					EvDisruptRefuge;
					pawns_on_line[col][line]=0;
				}
			}
		} else
			if (!EvCenterLine) {
				open_lines[EvSide]++;
				for (col=P_WHITE; col>=P_BLACK; col--)
					EvDisruptRefuge;
			}

	/*** evaluate king ***/
	for (col=P_WHITE; col>=P_BLACK; col--) {
		int kd=0;
		EvHeader((col ? "P_WHITE KING" : "P_BLACK KING"));
		Color=col;
		pos=g->sit.king[col];

		/* fill all the collected values into side_danger */
		for (line=D;line<=E;line++) {
			EvAddSideDanger("open line(s)",open_line_danger[open_lines[EvSide]]);
			EvAddSideDanger("half-open line(s)",halfopen_line_danger[halfopen_lines[col][EvSide]]);
			EvAddSideDanger("advanced sidepawns",sidepawn_advance_danger[advanced_sidepawns[col][EvSide]]);
		}

		line=Line(pos);

		/* if the baseline-squares on the a,b respectively the g,h lines are fully 'intact',
		   reduce the seriousness of the defects on the c,f lines if there is already
		   enough penalty. */	
		if(!defective_refuge[col][EvSide] && EvOnRefugeLine)
			EvAddSideDanger("intact refuge",KD_INTACT_REFUGE);
		if(side_danger[col][EvSide]<0)
			side_danger[col][EvSide]=0;

		/* specific mate threat: I call it "Anti-Fianchetto mate" */
		if( !(OwnPawns&mask64[ColCmp((EvSide?G2:B2),(EvSide?G7:B7))])
			 &&((EnemyPawns|EnemyBishops)&mask64[ColCmp((EvSide?F3:C3),(EvSide?F6:C6))])
			 &&(EnemyQueens) )
					EvAddSideDanger("anti fianchetto mate threat",KD_THREAT_ANTIFIANCHETTO);

		/* specific mate threat: The trojan horse mate */
		if( line>=E && EnemyRooks && EnemyQueens 
		    && !((g->sit.pawns[col]|g->sit.pawns[!col])&m_line[H]) )
			EvAddSideDanger("trojan horse mate threat",KD_THREAT_TROJANHORSE);

		/* convert the kd (king defect) unit into cp (centipawns) via conversion table */
		if(g->sit.castling_disallowed[col]==3) {
			if(EvSide==KINGSIDE)
				EvAddPosScore("total kingside danger",kd=-dangerousness_centipawns[side_danger[col][KINGSIDE]],-1) /* no ; here! */
			else
				EvAddPosScore("total queenside danger",kd=-dangerousness_centipawns[side_danger[col][QUEENSIDE]],-1);
		} else {
				/* only small castling possible */
				EvAddPosScore("total kingside danger (50%)",kd=-dangerousness_centipawns[side_danger[col][KINGSIDE]]>>1,-1) /* no ; here! */
				/* only great castling possible */
				EvAddPosScore("total queenside danger (50%)",kd=-dangerousness_centipawns[side_danger[col][QUEENSIDE]]>>1,-1);
			if( pos==ColCmp(E1,E8) && !((g->sit.pawns[col]|g->sit.pawns[!col])&m_line[E]) )
				EvAddPosScore("uncastled king and open E line",UNCASTLED_OPEN_E_LINE,-1);
		}
		/* if the enemy has no queen anymore, we may lessen the dangerousness penalty */
		if(kd && !EnemyQueens)
			EvAddPosScore("captured queen mitigation",-(kd>>1),-1);
		/* penalize loss of castling rights */
		if(!g->root_castling_disallowed[col])
			if(g->sit.castling_disallowed[col])
				if(line>C && line<G)
					/* this will never be shown in the static evaluation <.e> */
					EvAddPosScore("castling loss penalty",CASTLING_LOSS_PENALTY,-1);

		if(g->phase<ENDGAME)
			EvAddPosScore("midgame KING positional table",king_mid_posvalue[col][pos],-1);
	}

	/*** evaluate pawns ***/
	for(col=P_WHITE; col>=P_BLACK; col--) {
		EvHeader((col ? "P_WHITE PAWNS" : "P_BLACK PAWNS"));
		/* penalize multi-pawns (e.g. d4,d5) */
		EvAddPosScore("multi-pawn(s)",multipawn_penalty[col],-1);
		/* look at each pawn */
		remaining_figures=g->sit.pawns[col];
		for(num_pieces=0, pos_score_table=0, pos=0, pos_mask=BB1; remaining_figures && pos<64; num_pieces++) {
			int passed=0;
			while(!(pos_mask&remaining_figures)) {
				pos_mask<<=1;
				pos++;
			}
			ClearBit(remaining_figures,pos);
			/* positional table */
			pos_score_table+=pawn_posvalue[col][pos];
			
			/* penalize isolated pawns and isolated multi-pawns */
			if(! ( ((Line(pos)!=A ? m_line[Line(pos)-1] : 0)
			      | (Line(pos)!=H ? m_line[Line(pos)+1] : 0)) & g->sit.pawns[col]) )
				EvAddPosScore("isolated pawn",ISOLANI,pos); /* results in n*(MULTI_PAWN+ISOLANI) for n pawns */

			/* determine passed pawns (e.g. d4 and no enemy pawn in the rectangle c5-e5-e7-c7) */
			if(!(m_ranks_ahead[col][EvRelRank(pos)] & (EvSideLines | EvCurrentLine) & g->sit.pawns[!col])) {
				passedpawn_lines[col]|=mask8[Line(pos)];
				EvAddPosScore("passed pawn",passedpawn_bonus[EvRelRank(pos)],pos);
				passed=1;
				/* penalize blocked passed pawns */
				if(mask64[pos+EvColCmp(8,-8)]&g->sit.all_figures[!col])
					EvAddPosScore("blocked passed pawn",blockedpassedpawn_malus[EvRelRank(pos)],pos);
				/* give bonus for supporting king */
				if(EvRelRank(g->sit.king[col])>=EvRelRank(pos) && abs(Line(g->sit.king[col])-Line(pos))==1)
					EvAddPosScore("king's support for passed pawn",passedpawnwithking_bonus[EvRelRank(pos)],pos);
			} 

			/* determine two covered pawns (e.g. d4,e5 <-- e5, the more advanced one, will be seen here) */
			Color=!col;
			if(APawn(pos)&EnemyPawns)
				if(passed)
					if(EvRelRank(pos)==6)
						EvAddPosScore("covered passed pawn on 7th rank",COVERED_PASSED_7TH,pos) /* no ; here! */
					else
						EvAddPosScore("covered passed pawn",covered_bonus[EvRelRank(pos)],pos) /* no ; here! */
				else
					EvAddPosScore("covered pawn",covered_bonus[EvRelRank(pos)],pos);

			/* determine twin pawns (e.g. d4,e4) */
			if(EvSideSquares & g->sit.pawns[col])
				EvAddPosScore("twin pawn",twinpawn_bonus[EvRelRank(pos)],pos);
		}
		EvAddPosScore("PAWN positional table sum",pos_score_table,-1);
	}
	
	/*** evaluate knights ***/
	EvHeader("KNIGHTS");
	for(col=P_WHITE; col>=P_BLACK; col--) {
		remaining_figures=g->sit.knights[col];
		for(num_pieces=0, pos_score_table=0, pos=0, pos_mask=BB1; remaining_figures && pos<64; num_pieces++) {
			while(!(pos_mask&remaining_figures)) {
				pos_mask<<=1;
				pos++;
			}
			ClearBit(remaining_figures,pos);
			/* positional table */
			pos_score_table+=knight_posvalue[col][pos];			

			/* outposts */
			if(outpost_value[pos]) {
				Color=col;
				if(!(APawn(pos)&EnemyPawns)) {
					Color=!col;
					if(tmp_bb=APawn(pos)&EnemyPawns)
						EvAddPosScore("knight outpost",outpost_value[col][pos]*CntBits(tmp_bb),pos);
				}
			}
		}
		EvAddPosScore("KNIGHT positional table sum",pos_score_table,-1);
	}
	
	/*** evaluate bishops ***/
	EvHeader("BISHOPS");
	for(col=P_WHITE; col>=P_BLACK; col--) {
		remaining_figures=g->sit.bishops[col];
		for(num_pieces=0, pos_score_table=0, pos=0, pos_mask=BB1; remaining_figures && pos<64; num_pieces++) {
			while(!(pos_mask&remaining_figures)) {
				pos_mask<<=1;
				pos++;
			}
			ClearBit(remaining_figures,pos);
			/* positional table */
			pos_score_table+=bishop_posvalue[col][pos];

			/* outposts */
			if(outpost_value[pos]) {
				Color=col;
				if(!(APawn(pos)&EnemyPawns)) {
					Color=!col;
					if(tmp_bb=APawn(pos)&EnemyPawns)
						/* I personally consider a knight outpost more valuable than a bishop one, so i won't multiply. */
						EvAddPosScore("bishop outpost",outpost_value[col][pos],pos); 
				}
			}

			/* mobility */
			Color=col;
			EvAddPosScore("bishop mobility",CntBits(TBishop(pos)),pos);
			
			/* caged bishop */
			if (pos==EvColCmp(C1,C8) && CntBits(OwnPawns&bishop_cage_c[col])>=2 && !(OwnPawns&mask64[EvColCmp(C2,C7)])
			 || pos==EvColCmp(F1,F8) && CntBits(OwnPawns&bishop_cage_f[col])>=2 && !(OwnPawns&mask64[EvColCmp(F2,F7)]))
				EvAddPosScore("bishop caged on baseline",BISHOP_BASELINE_CAGED,pos);
		}

		/* twin bishops */
		if(num_pieces>1) EvAddPosScore("bishop pair",TWIN_BISHOPS,-1);

		/* fianchetto */
		Color=col;
		if( (Line(g->sit.king[col])>E && EvRelRank(g->sit.king[col])==0
		      && g->sit.pawns[col]&mask64[ColCmp(G3,G6)] && g->sit.bishops[col]&fianchetto_king_side[col])
		 || (Line(g->sit.king[col])<D && EvRelRank(g->sit.king[col])==0
		      && g->sit.pawns[col]&mask64[ColCmp(B3,B6)] && g->sit.bishops[col]&fianchetto_queen_side[col]) )
			EvAddPosScore("fianchettoed bishop",FIANCHETTO,Line(g->sit.king[col])<D ? ColCmp(B3,B6) : ColCmp (G3,G6));

		/* bishops have an advantage in endgames */
		if(g->phase==ENDGAME && num_pieces)
			EvAddPosScore("endgame bishop advantage",ENDGAME_BISHOP_BONUS*num_pieces,-1);
			
		EvAddPosScore("BISHOP positional table sum",pos_score_table,-1);
	}
	
	/*** evaluate rooks ***/
	EvHeader("ROOKS");
	for(col=P_WHITE; col>=P_BLACK; col--) {
		remaining_figures=g->sit.rooks[col];
		for(num_pieces=0, pos_score_table=0, pos=0, pos_mask=BB1; remaining_figures && pos<64; num_pieces++) {
			while(!(pos_mask&remaining_figures)) {
				pos_mask<<=1;
				pos++;
			}
			ClearBit(remaining_figures,pos);
			/* positional table */
			pos_score_table+=rook_posvalue[col][pos];			

			/* half-open and open lines */
			if(!pawns_on_line[col][Line(pos)])
				if(!pawns_on_line[!col][Line(pos)])
					EvAddPosScore("rook on open line",ROOK_OPEN_LINE,pos) /* no ; here! */
				else
					EvAddPosScore("rook on half-open line",ROOK_HALFOPEN_LINE,pos);

			/* rook on 7th/2nd rank */
			Color=col;
			if(Rank(pos)==ColCmp(6,1) && (ColCmp(g->sit.king[P_BLACK]>55,g->sit.king[P_WHITE]<8)
			                           || g->sit.pawns[!col]&m_rank[ColCmp(6,1)])) {	
				if(rank_attacks[pos][RankState(Rank(pos))]&(OwnQueens|OwnRooks))
					EvAddPosScore("reinforced rook on 7th/2nd rank",REINFORCED_ROOK_7TH_2ND,pos) /* no ; here! */
				else
					EvAddPosScore("single rook on 7th/2nd rank",SINGLE_ROOK_7TH_2ND,pos);
				/* evaluate the 7th/2nd 'absolute' - a term by A. Nimzowitsch, explained on p.63 in his book 'my system' */
				if( ColCmp(g->sit.king[P_BLACK]>55,g->sit.king[P_WHITE]<8)
				     && (rank_attacks[pos][RankState(Rank(pos))]|mask64[pos])&pawnline_tails_a
				     && (rank_attacks[pos][RankState(Rank(pos))]|mask64[pos])&pawnline_tails_h )
					EvAddPosScore("rook on 7th/2nd rank absolute",ROOK_7TH_2ND_ABSOLUTE,pos);
			}

			/* rook behind passed pawn: first, check if there's at least one passed pawn ahead. afterwards,    */	
			/* compare <line-attacks from 'pos'> with <the foremost pawn on the line containing a passed pawn> */
			if( passedpawn_lines[col]&mask8[Line(pos)] && ( line_attacks[pos][LineState(Line(pos))] 
			            & mask64[ EvRelFMB(g->sit.pawns[col]&m_line[Line(pos)]) ] ) )
				EvAddPosScore("rook supports passed pawn",PASSED_ROOK_SUPPORT,pos);			
		}
		EvAddPosScore("ROOK positional table sum",pos_score_table,-1);
	}

	/*** evaluate queens ***/
	EvHeader("QUEENS");
	for(col=P_WHITE; col>=P_BLACK; col--) {
		remaining_figures=g->sit.queens[col];
		for(num_pieces=0, pos_score_table=0, pos=0, pos_mask=BB1; remaining_figures && pos<64; num_pieces++) {
			while(!(pos_mask&remaining_figures)) {
				pos_mask<<=1;
				pos++;
			}
			ClearBit(remaining_figures,pos);

			/* positional table */
			pos_score_table+=queen_posvalue[col][pos];

			/* reinforcing a strong rook on the 7th/2nd rank? */
			Color=col;
			if(Rank(pos)==ColCmp(6,1) && (ColCmp(g->sit.king[P_BLACK]>55,g->sit.king[P_WHITE]<8)
			                           || g->sit.pawns[!col]&m_rank[ColCmp(6,1)])
			                          && rank_attacks[pos][RankState(Rank(pos))]&OwnRooks)
				EvAddPosScore("queen reinforces rook on 7th/2nd",REINFORCING_QUEEN_7TH_2ND,pos);
			
			/* if there are still undeveloped lightfigures, penalize for early queen development */
			if(lightfigure_startpositions[col]&g->sit.all_figures[col]
			    && !(g->sit.queens[col]&mask64[EvColCmp(D1,D8)]))
				EvAddPosScore("too early developed queen",PREMATURE_QUEEN,pos);
		}
		EvAddPosScore("QUEEN positional table sum",pos_score_table,-1);
	}

	/*** conclusion ***/
	score+=pos_score[P_WHITE] - pos_score[P_BLACK];
	EvHeader("CONCLUSION");
	if(print_score) {
		PHRIStaticEvaluationRow("total material balance",-1,P_BOTH,g->sit.material_score[P_WHITE]-g->sit.material_score[P_BLACK]);
		PHRIStaticEvaluationRow("total position balance",-1,P_BOTH,pos_score[P_WHITE]-pos_score[P_BLACK]);
		EvHeader("");
		PHRIStaticEvaluationRow("overall balance",-1,P_BOTH,score);
	}
	Color=color_save;
	
	g->seldepth=Max(g->seldepth,g->ply);
	return (Color ? -score : score);
}