
#include "const.h"
#include "types.h"
#include "globals.h"
#include "macros.h"
#include "proto.h"
#include "matrices.h"


/*********************************************************************************
* GENERATION OF NON-ATTACK MOVES: After calling this function, all new moves     *
*   might be found after the pointer *move.                                      *
*   The function itself returns the number of new moves.                         *
*********************************************************************************/
int GenerateN(GAME *g, MOVE *move, int move_offset) {
	BITBOARD remaining_figures=AllOwn;
	BITBOARD remaining_attacks;
	int src,dst,figure;
	int i;
	int move_type;
	int total_moves=move_offset;

	for(src=0;remaining_figures && src<64;src++)
		if(remaining_figures&mask64[src]) {
			remaining_attacks=0;
			figure=abs(Piece(src));
			switch (figure) {
				case PAWN:
					if(!ColCmp(TriCmp(47,src,56),TriCmp(7,src,16)))
						remaining_attacks=NPawn(src);
					break;
				case KNIGHT: remaining_attacks=NKnight(src); break;
				case BISHOP: remaining_attacks=NBishop(src); break;
				case ROOK: remaining_attacks=NRook(src); break;
				case QUEEN: remaining_attacks=NQueen(src); break;
				case KING: remaining_attacks=NKing(src);
					/* small castling */
					if(SmallCastlingPossible)
						move[total_moves++]=(CASTLING<<15)|(ColCmp(G1,G8)<<9)|(ColCmp(E1,E8)<<3)|KING;
					/* great castling */
					if(GreatCastlingPossible)
						move[total_moves++]=(CASTLING<<15)|(ColCmp(C1,C8)<<9)|(ColCmp(E1,E8)<<3)|KING;
					break;
				default: Error(errDetails,"unknown figure in GenerateN()!"); 
			}

			for(dst=0;remaining_attacks && dst<64;dst++)
				if(remaining_attacks&mask64[dst]) {
					move_type=NORMAL_MOVE;
					if(figure==PAWN) {
						if(IsDoubleStep(src,dst))
							move_type=DOUBLE_STEP;
						if(Rank(dst)==ColCmp(7,0)) {
							move_type=PROMOTION;
							/* for i==0, the promotion will be added later */
							for(i=3;i>0;i--)
								move[total_moves++]=(abs(Piece(dst))<<20)|(i<<18)|(move_type<<15)|(dst<<9)|(src<<3)|figure;							
						}
					}
					move[total_moves++]=(abs(Piece(dst))<<20)|(move_type<<15)|(dst<<9)|(src<<3)|figure;
					ClearBit(remaining_attacks,dst);
				}			
			ClearBit(remaining_figures,src);

		}
	return total_moves;
}


/*********************************************************************************
* GENERATION OF ATTACKS AND PROMOTIONS: After calling this function, all new     *
*   moves might be found after the pointer *move.                                *
*   The function itself returns the number of new moves.                         *
*********************************************************************************/
int GenerateAP(GAME *g, MOVE *move, int move_offset) {
	BITBOARD remaining_figures=AllOwn;
	BITBOARD remaining_attacks;
	int src,dst,figure;
	int i;
	int move_type;
	int total_moves=move_offset;

	for(src=0;remaining_figures && src<64;src++)
		if(remaining_figures&mask64[src]) {
			remaining_attacks=0;
			figure=abs(Piece(src));
			switch (figure) {
				case PAWN: remaining_attacks=APawn(src) & (AllEnemy | (g->sit.enpassant ? mask64[g->sit.enpassant] : 0));
					if(ColCmp(TriCmp(47,src,56),TriCmp(7,src,16)))
						remaining_attacks|=NPawn(src);
				break;
				case KNIGHT: remaining_attacks=AKnight(src) & AllEnemy; break;
				case BISHOP: remaining_attacks=ABishop(src) & AllEnemy; break;
				case ROOK: remaining_attacks=ARook(src) & AllEnemy; break;
				case QUEEN: remaining_attacks=AQueen(src) & AllEnemy; break;
				case KING: remaining_attacks=AKing(src) & AllEnemy; break;
				default:
					Error(errDetails,"unknown figure in GenerateAP()!");
					Dualprintf("\nexamining move: %c: %c%c->%c%c (t: %d/p: %d/c: %d)\n",pieceSymbol[figure],LineLiteral(src),RankLiteral(src),LineLiteral(dst),RankLiteral(dst),move_type,figure,Piece(dst));
					dPrintBitboard(AllOwn); Dualprintf("-");
					dPrintBitboard(OwnPawns); Dualprintf("-");
					dPrintBitboard(OwnKnights); Dualprintf("-");
					dPrintBitboard(OwnBishops); Dualprintf("-");
					dPrintBitboard(OwnQueens); Dualprintf("-");
					dPrintBitboard(OwnKing); Dualprintf("-");
			}

			for(dst=0;remaining_attacks && dst<64;dst++)
				if(remaining_attacks&mask64[dst]) {
					move_type=NORMAL_MOVE;
					if(abs(Piece(dst))==KING) 
						move_type=KING_CAPTURE;
					else /* this 'else' is important, since a promotion that captures a king is also a KING_CAPTURE! */
						if(figure==PAWN) {
							if(Piece(dst)==0)
								move_type=ENPASSANT_CAPTURE;
							if(Rank(dst)==ColCmp(7,0)) {
								move_type=PROMOTION;
								/* for i==0, the promotion will be added later */						
								for(i=3;i>0;i--)
									move[total_moves++]=(abs(Piece(dst))<<20)|(i<<18)|(move_type<<15)|(dst<<9)|(src<<3)|figure;
							}
						}
					move[total_moves++]=(abs(Piece(dst))<<20)|(move_type<<15)|(dst<<9)|(src<<3)|figure;
					ClearBit(remaining_attacks,dst);
				}			
			ClearBit(remaining_figures,src);

		}
	return total_moves;
}