

#include "const.h"
#include "types.h"
#include "globals.h"
#include "macros.h"
#include "proto.h"
#include "matrices.h"

/*********************************************************************************
* PROGRAM INITIALISATION: This routine is called directly at the beginning of    *
*   main(). The most interesting thing herein are the move generation routines.  *
*********************************************************************************/
void Init(GAME *g) {
	int i,j,tmp;
	unsigned char c=1,left,right;
	BITBOARD b=1,turtle_area;
	BITBOARD lu,rd; /* left up, right down */
	long l=1;

	/********************************************
	* INITIALIZE SEVERAL BITMASKS
	*/
	for (i=0;i<8;i++) {
		mask8[i]=c<<i;
		imask8[i]=~mask8[i];
	}
	for (i=0;i<32;i++) {
		mask32[i]=l<<i;
		imask32[i]=~mask32[i];
	}
	for (i=0;i<64;i++) {
		mask64[i]=b<<i;
		imask64[i]=~mask64[i];
	}

	/* line / rank bitmasks */
	m_line[0]=0;
	m_rank[0]=0;
	for (i=0;i<8;i++) {
		m_line[i]=0;
		m_rank[i]=0;
		for (j=i*8;j<(i+1)*8;j++)
			m_rank[i]|=mask64[j];
		for (j=i;j<64;j+=8)
			m_line[i]|=mask64[j];
	}

	/* all ranks ahead a certain rank, e.g. set 7th&8th rank to be ahead the 6th rank for white */
	for (i=0;i<8;i++) {
		for (j=1+i;j<8;j++)
			m_ranks_ahead[P_WHITE][i]|=m_rank[j];
		for (j=6-i;j>=0;j--)
			m_ranks_ahead[P_BLACK][i]|=m_rank[j];
	}


	/********************************************
	* INITIALIZE EVALUATION MATRICES AND THE kd<->cp CONVERTER (king defects <--> centipawns)
	*/
	for (i=0;i<8;i++)
		for (j=0;j<8;j++) {
			/* copy all the black 'sailing-boat' arrays flipped into the white ones */
			knight_posvalue[P_WHITE][SqVal(i,7-j)]=knight_posvalue[P_BLACK][SqVal(i,j)];
			bishop_posvalue[P_WHITE][SqVal(i,7-j)]=bishop_posvalue[P_BLACK][SqVal(i,j)];
			rook_posvalue[P_WHITE][SqVal(i,7-j)]=rook_posvalue[P_BLACK][SqVal(i,j)];
			queen_posvalue[P_WHITE][SqVal(i,7-j)]=queen_posvalue[P_BLACK][SqVal(i,j)];
			pawn_posvalue[P_WHITE][SqVal(i,7-j)]=pawn_posvalue[P_BLACK][SqVal(i,j)];
			king_mid_posvalue[P_WHITE][SqVal(i,7-j)]=king_mid_posvalue[P_BLACK][SqVal(i,j)];
			outpost_value[P_WHITE][SqVal(i,7-j)]=outpost_value[P_BLACK][SqVal(i,j)];
		}
	
	for (i=0;i<=100;i++)
		dangerousness_centipawns[i]=(int) Min(king_safety_care_1st*log(i*king_safety_care_2nd/100.0+1),king_safety_upper_limit);
	
	/********************************************
	* SET UP SEVERAL BITBOARDS ON THE CHESS BOARD
	*/
	castlingfields_king_side[P_WHITE]=mask64[F1]|mask64[G1];	
	castlingfields_king_side[P_BLACK]=mask64[F8]|mask64[G8];
	castlingfields_queen_side[P_WHITE]=mask64[D1]|mask64[C1]|mask64[B1];
	castlingfields_queen_side[P_BLACK]=mask64[D8]|mask64[C8]|mask64[B8];
	fianchetto_king_side[P_WHITE]=mask64[F1]|mask64[G2]|mask64[H1];
	fianchetto_king_side[P_BLACK]=mask64[F8]|mask64[G7]|mask64[H8];
	fianchetto_queen_side[P_WHITE]=mask64[A1]|mask64[B2]|mask64[C1];
	fianchetto_queen_side[P_BLACK]=mask64[A8]|mask64[B7]|mask64[C8];
	
	lightfigure_startpositions[P_WHITE]=mask64[B1]|mask64[C1]|mask64[F1]|mask64[G1];
	lightfigure_startpositions[P_BLACK]=mask64[B8]|mask64[C8]|mask64[F8]|mask64[G8];
	
	pawnlines=m_rank[1]|m_rank[6];
	pawnline_tails_a=pawnlines&m_line[A];
	pawnline_tails_h=pawnlines&m_line[H];
	
	bishop_cage_c[P_WHITE]=mask64[B2]|mask64[D2];
	bishop_cage_c[P_BLACK]=mask64[B7]|mask64[D7];
	bishop_cage_f[P_WHITE]=mask64[E2]|mask64[G2];
	bishop_cage_f[P_BLACK]=mask64[E7]|mask64[G7];

	
	/********************************************
	* PRE-CALCULATE THE KNIGHT MOVES
	*/
	for (i=0;i<64;i++) {
		knight_attacks[i]=0;
		for (j=0;j<64;j++)
			if (((abs(Rank(i)-Rank(j))==2) && (abs(Line(i)-Line(j))==1)) ||
				  ((abs(Rank(i)-Rank(j))==1) && (abs(Line(i)-Line(j))==2)))
				knight_attacks[i]|=mask64[j];
				/* simple brute-force method, I'm not really creative at the moment. :) */
	}

	/********************************************
	* PRE-CALCULATE THE KING MOVES
	*/
	for (i=0;i<64;i++) {
		king_attacks[i]=0;
		for (j=0;j<64;j++)
			if (abs(Rank(i)-Rank(j))<=1 && abs(Line(i)-Line(j))<=1 && i!=j)
				king_attacks[i]|=mask64[j];
	}

	/********************************************
	* PRE-CALCULATE THE PAWN MOVES
	*/
	for (i=0;i<64;i++)
		for (j=P_BLACK;j<=P_WHITE;j++) {
			pawn_attacks[i][j]=0;
			pawn_nonattacks_singlestep[i][j]=0;
			if((j==P_WHITE && i>=0 && i<56) || (j==P_BLACK && i>7 && i<64)) {
				/* for (j==P_WHITE), the second value will be added */
				pawn_nonattacks_singlestep[i][j]|=mask64[i-8+j*16];
				if(Line(i)!=A) pawn_attacks[i][j]|=mask64[i-9+j*16];
				if(Line(i)!=H) pawn_attacks[i][j]|=mask64[i-7+j*16];
			}
		}
	

	/********************************************
	* PRE-CALCULATE THE SLIDING PIECES (ROOKS, BISHOPS, QUEENS) MOVES
	*/
	for (i=0;i<64;i++)
		for (j=0;j<256;j++) {
			/* now we place a bit-sized turtle and let him run across the rank. 'b' is his initial position. */
			b=BB1<<i;
			
			/* *************** */
			/* 1. rank attacks */
			/* *************** */
			turtle_area=((BITBOARD) j << slide_shift[i]);

			/* put the turtle on the board, looking LEFT. (shift right -> decrease lu -> walk left) */
			if (Line(i)>0) {
				lu=b;
				/* let the turtle continue, then check whether it's still in his area and whether it's still on the same rank. */
				while ( !((lu>>=1)&(BB1 << slide_shift[i])) && !(lu&turtle_area) )
					lu|=b; /* <- if the turtle hasn't collided yet with a bit. */
			} else
				lu=0;
					
			/* put the turtle on the board, looking RIGHT. (shift left -> increase rd -> walk right) */
			if (Line(i)<7) {
				rd=b; 
				/* let the turtle continue, then check if it's still in his area and if it's still on the same rank. */
				while ( !((rd<<=1)&(BB1 << (slide_shift[i]+7))) && !(rd&((BITBOARD) j << slide_shift[i])) )
					rd|=b; /* <- if the turtle hasn't collided yet with a bit. */
			} else
				rd=0;
	
			rank_attacks[i][j]=(lu|rd);
			/* we throw the turtle onto the graveyard and create a new one. */


			/* *************** */
			/* 2. line attacks */
			/* *************** */
			turtle_area=((mask8[0]&j ? (mask64[   Line(i)]) : 0)
			           | (mask8[1]&j ? (mask64[ 8+Line(i)]) : 0)
			           | (mask8[2]&j ? (mask64[16+Line(i)]) : 0)
			           | (mask8[3]&j ? (mask64[24+Line(i)]) : 0)
			           | (mask8[4]&j ? (mask64[32+Line(i)]) : 0)
			           | (mask8[5]&j ? (mask64[40+Line(i)]) : 0)
			           | (mask8[6]&j ? (mask64[48+Line(i)]) : 0)
			           | (mask8[7]&j ? (mask64[56+Line(i)]) : 0));
		
			/* put the turtle on the board, looking LEFT. (shift right -> decrease lu -> walk upwards) */
			if (Rank(i)>0) {
				lu=b;
				/* let the turtle continue, then check if it's still in his area and if it's still on the same line. */
				while ( !((lu>>=8)&m_rank[0]) && !(lu&turtle_area) )
					lu|=b; /* <- if the turtle hasn't collided yet with a bit. */
			} else
				lu=0;
		
			/* put the turtle on the board, looking RIGHT. (shift left -> increase rd -> walk downwards) */
			if (Rank(i)<7) {
				rd=b; 
				/* let the turtle continue, then check if it's still in his area and if it's still on the same line. */
				while ( !((rd<<=8)&m_rank[7]) && !(rd&turtle_area) )
					rd|=b; /* <- if the turtle hasn't collided yet with a bit. */
			} else
				rd=0;
	
			line_attacks[i][j]=(lu|rd);


			/* *************************************** */
			/* 3. diagonals from left to right attacks */
			/* *************************************** */
			turtle_area=((mask8[0]&j && (diagLR_length[i]>=1) ? (mask64[diagLR_square[0][i]]) : 0)
			           | (mask8[1]&j && (diagLR_length[i]>=2) ? (mask64[diagLR_square[1][i]]) : 0)
			           | (mask8[2]&j && (diagLR_length[i]>=3) ? (mask64[diagLR_square[2][i]]) : 0)
			           | (mask8[3]&j && (diagLR_length[i]>=4) ? (mask64[diagLR_square[3][i]]) : 0)
			           | (mask8[4]&j && (diagLR_length[i]>=5) ? (mask64[diagLR_square[4][i]]) : 0)
			           | (mask8[5]&j && (diagLR_length[i]>=6) ? (mask64[diagLR_square[5][i]]) : 0)
			           | (mask8[6]&j && (diagLR_length[i]>=7) ? (mask64[diagLR_square[6][i]]) : 0)
			           | (mask8[7]&j && (diagLR_length[i]>=8) ? (mask64[diagLR_square[7][i]]) : 0));

			/* put the turtle on the board, looking LEFT. (shift right -> decrease lu -> walk upwards leftwards) */
			if (Rank(i)>0 && Line(i)>0) {
				lu=b;
				/* let the turtle continue, then check if it's still in his area and if it's still on the same line. */
				while ( !((lu>>=9)&(m_rank[0]|m_line[0])) && !(lu&turtle_area) )
					lu|=b; /* <- if the turtle hasn't collided yet with a bit. */
			} else
				lu=0;
		
			/* put the turtle on the board, looking RIGHT. (shift left -> increase rd -> walk downwards rightwards) */
			if (Rank(i)<7 && Line(i)<7) {
				rd=b; 
				/* let the turtle continue, then check if it's still in his area and if it's still on the same line. */
				while ( !((rd<<=9)&(m_rank[7]|m_line[7])) && !(rd&turtle_area) )
					rd|=b; /* <- if the turtle hasn't collided yet with a bit. */
			} else
				rd=0;
	
			diagLR_attacks[i][j]=(lu|rd);


			/* *************************************** */
			/* 4. diagonals from right to left attacks */
			/* *************************************** */
			turtle_area=((mask8[0]&j && (diagRL_length[i]>=1) ? (mask64[diagRL_square[0][i]]) : 0)
			           | (mask8[1]&j && (diagRL_length[i]>=2) ? (mask64[diagRL_square[1][i]]) : 0)
			           | (mask8[2]&j && (diagRL_length[i]>=3) ? (mask64[diagRL_square[2][i]]) : 0)
			           | (mask8[3]&j && (diagRL_length[i]>=4) ? (mask64[diagRL_square[3][i]]) : 0)
			           | (mask8[4]&j && (diagRL_length[i]>=5) ? (mask64[diagRL_square[4][i]]) : 0)
			           | (mask8[5]&j && (diagRL_length[i]>=6) ? (mask64[diagRL_square[5][i]]) : 0)
			           | (mask8[6]&j && (diagRL_length[i]>=7) ? (mask64[diagRL_square[6][i]]) : 0)
			           | (mask8[7]&j && (diagRL_length[i]>=8) ? (mask64[diagRL_square[7][i]]) : 0));

			/* put the turtle on the board, looking LEFT. (shift right -> decrease lu -> walk upwards rightwards) */
			if (Rank(i)>0 && Line(i)<7) {
				lu=b;
				/* let the turtle continue, then check if it's still in his area and if it's still on the same line. */
				while ( !((lu>>=7)&(m_rank[0]|m_line[7])) && !(lu&turtle_area) )
					lu|=b; /* <- if the turtle hasn't collided yet with a bit. */
			} else
				lu=0;
		
			/* put the turtle on the board, looking RIGHT. (shift left -> increase rd -> walk downwards leftwards) */
			if (Rank(i)<7 && Line(i)>0) {
				rd=b; 
				/* let the turtle continue, then check if it's still in his area and if it's still on the same line. */
				while ( !((rd<<=7)&(m_rank[7]|m_line[0])) && !(rd&turtle_area) )
					rd|=b; /* <- if the turtle hasn't collided yet with a bit. */
			} else
				rd=0;
	
			diagRL_attacks[i][j]=(lu|rd);		
		}


#ifdef BCC32
	/********************************************
	* INITIALIZE THE MICROSOFT EXTRAWURSCHT (sorry, no english translation for this) CONCERNING I/O PIPES
	*/
	win32_inputhandle = GetStdHandle(STD_INPUT_HANDLE);
	win32_pipe = !GetConsoleMode(win32_inputhandle, &win32_consolemode);
	if (!win32_pipe) {
		SetConsoleMode(win32_inputhandle, win32_consolemode & ~(ENABLE_MOUSE_INPUT|ENABLE_WINDOW_INPUT));
		FlushConsoleInputBuffer(win32_inputhandle);
	}
#endif

	/********************************************
	* --> OTHER INIT ROUTINES
	*/
	transposition_table=(HASHENTRY*) malloc(1);
	InitTranspositionTable(hash_megabytes);
	InitZobristKeys();
	InitOriginalPosition(g);
}


/*********************************************************************************
* INITIALIZE ORIGINAL POSITION                                                   *
*********************************************************************************/
void InitOriginalPosition(GAME *g) {
	/* empty g and set up the intial chess board */
	memset(g,0,sizeof(GAME));
	memcpy(&(g->sit.square),initial_sq,sizeof(initial_sq));
	/* initialize some variables in g */
	g->sit.color_to_move=P_WHITE;
	g->result=RESULT_OPEN;
	g->fullmoves_number=1;
	g->ply=1;
	/* update all bitboards with the initial position */
	RefreshBitboards(g);
	UpdateHalfmoveIndex;
}


/*********************************************************************************
* INITIALIZE ZOBRIST KEYS                                                        *
*********************************************************************************/
void InitZobristKeys(void) {
	int figure, color, square, forbiddance;
	
	for (square=0;square<64;square++) {
		for (color=0;color<2;color++) {
			/* zobrist keys for each figure on each square */
			for (figure=1;figure<7;figure++)		
				zobrist_figure[color][figure][square]=Random64();
			/* zobrist keys for castling rights */
			for (forbiddance=0;forbiddance<4;forbiddance++)
				zobrist_castling_disallowed[color][forbiddance]=Random64();
		}
		/* zobrist keys for the en passant square */
		zobrist_enpassant[square]=Random64();
	}
	/* zobrist keys for the current color */
	zobrist_whitetomove=Random64();
}


/*********************************************************************************
* INITIALIZE TRANSPOSITION TABLE                                                 *
*********************************************************************************/
void InitTranspositionTable(int megabytes) {
	long hash_bytes=(long) megabytes*1048576;
	
	if (transposition_table=(HASHENTRY*) realloc(transposition_table,hash_bytes))
		; /*memset(transposition_table,0,hash_bytes);*/
	else {
		Error(errDetails,"Not enough memory!");
		exit(EXIT_FAILURE);
	}
	tt_entries=hash_bytes/sizeof(HASHENTRY);
}