
#include "const.h"
#include "types.h"
#include "globals.h"
#include "macros.h"
#include "proto.h"
#include "matrices.h"


/*********************************************************************************
* PARSE POSSIBLE INTERFACE CHANGES: The user might switch between UCI and PHRI.  *
*********************************************************************************/
int ParseInterfaceChange (char *buffer) {
	int previous_chess_interface=chess_interface;
	
	if (!strcmp(buffer,"uci")) {
		UCIIdentify();
		chess_interface=UCI;
	}
	if (!strcmp(buffer,"phri"))
		chess_interface=PHRI;
	
	return (chess_interface!=previous_chess_interface);
}


/*********************************************************************************
* PARSE MOVESTRING: This function parses moves like Sf3, e5, Lgxf6, cxd8=T, d3d4 *
*   My first self-written function in C, somewhen in the beginning of 2003. :)   *
*********************************************************************************/
MOVE ParseMove (GAME *g, char *buffer) {
	const char figures[21]="SsNnLlBbTtRrQqQqKkKk";
	const char lines[17]="AaBbCcDdEeFfGgHh";
	const char ranks[9]="12345678";

	unsigned char srcL=0, srcR=0, src;
	unsigned char dstL, dstR, dst;
	signed char figure=0, promotion_figure=QUEEN, move_type=NORMAL_MOVE;
	int i,j;
	char *s, *last;
	BITBOARD possible_origins=0;

	last=buffer+strlen(buffer)-1;
	if (last<buffer) return 0;

	/*******************
	* PROMOTION CUTOFF *
	********************/ 
	if (s=strchr(figures,*last)) {
		promotion_figure=(s-figures)/4+2;
		last--;
	}
		
	/********************
	* CHECK/MATE CUTOFF *
	********************/ 
	if (s=strchr(buffer,'='))
		last--;
	if (s=strchr(buffer,'+'))
		last--;
	if (s=strchr(buffer,'#'))
		last--;
	
	/*****************
	* CASTLING CHECK *
	*****************/ 
	if (!strcmp(buffer,"o-o") || !strcmp(buffer,"0-0") || !strcmp(buffer,"O-O"))
		if(SmallCastlingPossible) {
			return (CASTLING<<15)|(ColCmp(G1,G8)<<9)|(ColCmp(E1,E8)<<3)|KING;
		} else {
			Error(errDetails,"Small Castling not allowed, alreadymoved-flag = %d.\n",g->sit.castling_disallowed[Color]);
			return 0;
		}
			
	if (!strcmp(buffer,"o-o-o") || !strcmp(buffer,"0-0-0") || !strcmp(buffer,"O-O-O"))
		if(GreatCastlingPossible) {
			return (CASTLING<<15)|(ColCmp(C1,C8)<<9)|(ColCmp(E1,E8)<<3)|KING;
		} else {
			Error(errDetails,"Great Castling not allowed, alreadymoved-flag = %d.\n",g->sit.castling_disallowed[Color]);
			return 0;
		}

					
	/*******************
	* DESTINATION RANK *
	*******************/ 
	if (s=strchr(ranks,*last))
		dstR=s-ranks | 128;
	else return 0;
	last--;
	if (last<buffer) return 0;

	/*******************
	* DESTINATION LINE *
	*******************/ 
	if (s=strchr(lines,*last))
		dstL=(s-lines)/2 | 128;
	else return 0;
	last--;
	
	/* at this point, next char can be:
	 * 
	 * 1) not existent (-> pawn move)
	 * 2) "x" or ":" indicator (-> capture)
	 * 3) "-" indicator (-> long notation)
	 * 4) source line
	 * 5) source rank
	 * 6) figure indicator
	 *
	 */
	
	/* case 1 */
	if (last<buffer) {
		figure=PAWN;
		goto revise_move; /* determination finished */
	}
	
	/* cases 2 and 3 */
	if ((*last=='x') || (*last==':') || (*last=='-')) {
		last--;
		if (last<buffer) return 0;
	}
	
	/* case 4 */
	if (s=strchr(ranks,*last)) {
		srcR=s-ranks | 128;
		last--;
		if (last<buffer)
			goto revise_move; /* determination finished */
	}

	/* case 5 */
	if (s=strchr(lines,*last)) {
		srcL=(s-lines)/2 | 128;
		last--;
		if (last<buffer) {
			if (!srcR) /* this way, it's a pawn capture, like 'exf4' */
				figure=PAWN;			
			goto revise_move; /* determination finished */
		}
	}
	
	/* case 6 */
	if (s=strchr(figures,*last)) {
		figure=(s-figures)/4+2;
		last--;
		if (last<buffer)
			goto revise_move; /* determination finished */
	}		
	
	/*************************
	* PHASE 2: MOVE REVISION *
	*************************/
	revise_move:
	dst=SqVal((dstL&=127),(dstR&=127));

	/* complete src information already supplied in phase 1? */
	if (srcL && srcR) {
		src=SqVal(srcL&127,srcR&127);
		
		/* catch UCI style moves (e.g. g1f3) */
		if (!figure) figure=abs(Piece(src));

		/* catch UCI style moves with wrong preceding figure indicator (e.g. Bg1f3 at move 1). */
		if (figure!=abs(Piece(src))) return 0;
	}

	promotion_figure-=2; /* because promotion_figure has to be 2-bit, we scale it down from 2..5 to 0..3 */
		
	/* NOW: figure and dst are known. */
	switch (figure) {
		case 0:      return 0;
		case PAWN:   possible_origins=OwnPawns;   break;
		case KNIGHT: possible_origins=OwnKnights; break;
		case BISHOP: possible_origins=OwnBishops; break;
		case ROOK:   possible_origins=OwnRooks;   break;
		case QUEEN:  possible_origins=OwnQueens;  break;
		case KING:   possible_origins=OwnKing;    break;
	}

	if(srcL) possible_origins&=m_line[srcL&127];
	if(srcR) possible_origins&=m_rank[srcR&127];
	
	for(i=0;possible_origins && i<64;i++) /* almost equal to GenerateN() in generate.c */
		if(possible_origins&mask64[i])
			switch (figure) {
				case 0: 
					Error(errDetails,"undetermined figure!");
					return 0;
				case PAWN:   if (!(TPawn(i)&mask64[dst]))    { ClearBit(possible_origins,i); } break;
				case KNIGHT: if (!(TKnight(i)&mask64[dst]))  { ClearBit(possible_origins,i); } break;
				case BISHOP: if (!(TBishop(i)&mask64[dst]))  { ClearBit(possible_origins,i); } break;
				case ROOK:   if (!(TRook(i)&mask64[dst]))    { ClearBit(possible_origins,i); } break;
				case QUEEN:  if (!(TQueen(i)&mask64[dst]))   { ClearBit(possible_origins,i); } break;
				case KING:   if (!(TKing(i)&mask64[dst]))    { ClearBit(possible_origins,i); } break;
				default: break;
			}
	
	if (CntBits(possible_origins)==1) {
		src=LSB(possible_origins);
		if (figure==KING && abs(Line(src)-Line(dst))==2)
			move_type=CASTLING;
		if (figure==PAWN) {
			/* double step? */
			if(IsDoubleStep(src,dst))
				move_type=DOUBLE_STEP;
			/* en passant capture? */
			if(Piece(dst)==0 && Line(src)!=Line(dst))
				move_type=ENPASSANT_CAPTURE;
			/* promotion? */
			if (dstR==ColCmp(7,0)) {
				move_type=PROMOTION;			
				if (!TriCmp(-1,promotion_figure,4)) {
					Error(errDetails,"Illegal Promotion!");
					return 0;
				}
			}
		}
		return (abs(Piece(dst))<<20)|(promotion_figure<<18)|(move_type<<15)|(dst<<9)|(src<<3)|figure;;
	} else {
		if (CntBits(possible_origins)>1) {
			Error(errDetails,"Unclear Move:\n");
			dPrintBitboard(possible_origins);
		}
		if (CntBits(possible_origins)==0)
			Error(errDetails,"Impossible Move.");
		return 0;
	}
}


/*********************************************************************************
* PARSE FORSYTH-EDWARDS NOTATION (FEN): Refer to the documentation for details   *
*********************************************************************************/
char* ParseFEN (GAME *g, char *str) {
	const char figures[13]="PpNnBbRrQqKk";
	const char line_letters[9]="abcdefgh";
	const char rank_numbers[9]="12345678";
	const char delim[2]=" ";
	char *tok, *s;
	char line,rank;
	int len;
	
	len=strlen(str);
	memset(g->sit.square,0,sizeof(g->sit.square));
	
	/* first token: the pieces themselves */
	tok=strtok(str,delim);
	rank=7; line=0;
	do {
		if (*tok=='/') {
			rank--;
			line=0;
		}
		if (s=strchr(figures,*tok)) {
			Piece(SqVal(line,rank))=(s-figures)/2+1;
			if((s-figures)%2)
				Piece(SqVal(line,rank))*=-1;
			line++;
		}
		if (s=strchr(rank_numbers,*tok))
			line+=s-rank_numbers+1;
	} while(*(++tok));

	/* 2nd token: player to move */
	tok=strtok(NULL,delim);
	if (*tok=='w')
		Color=1;
	else if (*tok=='b')
		Color=0;
	else
		Error(errDetails,"2nd token!");
		
	/* 3nd token: castling availibility */
	tok=strtok(NULL,delim);
	g->sit.castling_disallowed[P_BLACK]=3;
	g->sit.castling_disallowed[P_WHITE]=3;
	do {
		if (*tok=='K') g->sit.castling_disallowed[P_WHITE]&=imask8[0];
		if (*tok=='Q') g->sit.castling_disallowed[P_WHITE]&=imask8[1];
		if (*tok=='k') g->sit.castling_disallowed[P_BLACK]&=imask8[0];
		if (*tok=='q') g->sit.castling_disallowed[P_BLACK]&=imask8[1];	
	} while (*(++tok));
	
	/* 4th token: en passant */
	tok=strtok(NULL,delim);
	if (*tok!='-') {
		if (s=strchr(line_letters,*tok))
			line=s-line_letters;
		else
			Error(errDetails,"4nd token / line!");
		if (s=strchr(rank_numbers,*(++tok)))
			rank=s-rank_numbers;
		else
			Error(errDetails,"4nd token / rank!");
		g->sit.enpassant=SqVal(line,rank);
	} else
		g->sit.enpassant=0;
	
	/* 5th token: halfmove counter for 50move rule */
	tok=strtok(NULL,delim);
	g->sit.rule50counter=atoi(tok);
	
	/* 6th token: fullmove counter */
	tok=strtok(NULL,delim);
	g->fullmoves_number=atoi(tok);

	RefreshBitboards(g);

	s=strrchr(tok,'\0')+1;
	if(s>str+len)
		return 0;
	else
		return s;
}