
#include "const.h"
#include "types.h"
#include "globals.h"
#include "macros.h"
#include "proto.h"
#include "matrices.h"


/*********************************************************************************
* 64-BIT RANDOM                                                                  *
*********************************************************************************/
BITBOARD Random64(void) {
	/* since RAND_MAX varies from compiler to compiler, we assume a minimum size of 15 bits */
	return rand() ^ ((BITBOARD) rand()<<15) 
	              ^ ((BITBOARD) rand()<<30)
	              ^ ((BITBOARD) rand()<<45)
	              ^ ((BITBOARD) rand()<<60);
}



/*********************************************************************************
* REFRESHBITBOARDS                                                               *
*********************************************************************************/
int RefreshBitboards(GAME *g) {
	int sq;
	char cSave=Color;
	
	for(Color=P_WHITE; Color>=P_BLACK; Color--) {
		OwnPawns=0;
		OwnKnights=0;
		OwnBishops=0;
		OwnRooks=0;
		OwnQueens=0;
		AllOwn=0;
		OwnScore=0;
		g->sit.pawn_count[Color]=0;
	}
	AllFigures=0;
	g->sit.zobrist=0;
	for (sq=0;sq<64;sq++) {
		Color=(Piece(sq)>0);
		if (abs(Piece(sq))) {
			g->sit.all_figures[Color]|=mask64[sq];
			g->sit.all_figures[P_BOTH]|=mask64[sq];
			switch (abs(Piece(sq))) {
				case PAWN:   g->sit.pawns[Color]|=mask64[sq];   
				             g->sit.pawn_count[Color]++;        break;
				case KNIGHT: g->sit.knights[Color]|=mask64[sq]; break;
				case BISHOP: g->sit.bishops[Color]|=mask64[sq]; break;
				case ROOK:   g->sit.rooks[Color]|=mask64[sq];   break;
				case QUEEN:  g->sit.queens[Color]|=mask64[sq];  break;
				case KING:   g->sit.king[Color]=sq; 
					XORZobristOwnKing(sq);
				break;
			}
			g->sit.zobrist^=zobrist_figure[Color][abs(Piece(sq))][sq];
			if(abs(Piece(sq))!=KING)
				g->sit.material_score[Color]+=piece_value[abs(Piece(sq))];
		}
	}
	
	if(Color==P_WHITE)
		XORZobristWTM;
	XORZobristCastling;
	SwapColors; XORZobristCastling; SwapColors;
	if(g->sit.enpassant)
		XORZobristEP(g->sit.enpassant);
	Color=cSave;
	return 1;
}


/*********************************************************************************
* ATTACKEDBYENEMY                                                                *
*********************************************************************************/
int AttackedByEnemy(GAME *g, int sq) {
	if (APawn(sq) & EnemyPawns) return 1;
	if (AKnight(sq) & EnemyKnights) return 1;
	if (ABishop(sq) & EnemyBishops) return 1;
	if (ARook(sq) & EnemyRooks) return 1;
	if (AQueen(sq) & EnemyQueens) return 1;
	if (AKing(sq) & EnemyKing) return 1;
	return 0;
}


/*********************************************************************************
* DETECTENDOFGAME                                                                *
*********************************************************************************/
int DetectEndOfGame(GAME *g) {
	if(g->result!=RESULT_OPEN) {
		switch(g->result) {
			case RESULT_DRAW_STALE: Dualprintf("\nThis Game ends in a draw due to stalemate!"); break;
			case RESULT_DRAW_MATERIAL: Dualprintf("\nThis Game ends in a draw for lack of mating material!"); break;
			case RESULT_DRAW_REPETITION: Dualprintf("\nThis Game ends in a draw due to repetition!"); break;
			case RESULT_DRAW_50MOVE: Dualprintf("\nThis Game ends in a draw for lack of captures or pawn moves in the last 50 moves!"); break;
			case RESULT_MATED_WHITE: Dualprintf("\nWhite checkmated! Congratulations, black player!"); break;
			case RESULT_MATED_BLACK: Dualprintf("\nBlack checkmated! Congratulations, white player!"); break;
		}
		Dualprintf("\n\n\n\nLet's play another game!");
		InitOriginalPosition(g);
		return 1;
	}
	return 0;
}


/*********************************************************************************
* DETECT DRAW                                                                    *
*********************************************************************************/
int DetectDraw(GAME *g, int *value, int move_played) {
	int i;
	int count;
	/* stalemate is handled externally in all the exploration routines */
	
	/* draw by lack of material */
	for(i=P_BLACK, count=0; i<=P_WHITE; i++)
		if(!g->sit.pawn_count[i] && g->sit.material_score[i]<=300)
			count++;
	
	if (count>=2) {
		*value=0;
		if(move_played)
			g->result=RESULT_DRAW_MATERIAL;
		return 1;
	}
	
	/* draw by 3x repetition */
	for(i=g->sit.last_irreversible_move, count=0; i<=g->halfmove_index+g->ply-1; i++) {
		if(g->sit.zobrist==g->passed_zobrist[i])
			count++;
	}
	if (count>=3) {
		*value=0;
		if(move_played)
			g->result=RESULT_DRAW_REPETITION;
		return 1;
	}
	
	/* draw by 50 move counter */
	if(g->sit.rule50counter>=100) {
		*value=0;
		if(move_played)
			g->result=RESULT_DRAW_50MOVE;
		return 1;
	}
	return 0;
}


/*********************************************************************************
* DETERMINE CURRENT GAME PHASE                                                   *
*********************************************************************************/
int DeterminePhase(GAME *g) {
	int col;
	
	/* detect opening */
	if(g->halfmove_index<42)
		return OPENING;
	for(col=P_BLACK; col<=P_WHITE; col++)
		if(lightfigure_startpositions[col]&g->sit.all_figures[col])
			return OPENING;
			
	/* detect endgame */
	if(g->sit.material_score[P_WHITE]<2000 && g->sit.material_score[P_BLACK]<2000)
		return ENDGAME;
		
	/* in any other case, we're in the middle game */
	return MIDDLEGAME;
}