
#pragma once



#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>

/*********************************************************************************
* TYPE DEFINITIONS                                                               *
*********************************************************************************/

/*********************************************************************************
 * defintion of 64-bit variables depends on the current OS
 */
#ifdef BCC32
	typedef unsigned __int64 BITBOARD; /* MMX optimized */
#else
	typedef unsigned long long BITBOARD; /* ANSI C99 */
#endif

/*********************************************************************************
 * MOVE has 32 bits:
 *  xxx xxxxxx xxxxxx xxx xx xxx x xxxxxxxx
 * |-A-|--B---|--C---|-D-|E-|-F-|G|--------|
 *
 * A: 3 bits: piece type (0..7)
 * B: 6 bits: source square (0..63)
 * C: 6 bits: destination square (0..63)
 * D: 3 bits: move type: see enum 'move_types'
 * E: 2 bits: promotion piece:
 *            0 knight
 *            1 bishop
 *            2 rook
 *            3 queen
 * F: 3 bits: captured piece
 * G: 1 bit: mate flag
 * 
 *   the rest is currently unused.
 */
typedef long MOVE;


/*********************************************************************************
 * SITBACKUP is used in DoMove() and UnDoMove() to restore the position before
 * DoMove() was called.
 */
typedef struct {
	char move_type;
	char src, dst;
	char src_figure;
	char dst_figure;
	char enpassant;
	int captured;
	int rule50counter;
	int last_irreversible_move;
	char castling_disallowed[2];
} SITBACKUP;


/*********************************************************************************
 * HASH ENTRY: First, we store the full 64-bit Zobrist key into <verification_key>
 * to verify that we're examining the right position.
 *
 * After that, we store all the information we need about the position in a 64-bit
 * variable called <data>:
 *
 *  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx xx   x xxxxxxxxxxxxxxx xxxxxxxxxxxxxxxx
 * |--------------A---------------|B-| |C|-------D-------|--------E-------|
 *
 * A: 30 bits: best move in this situation (type MOVE, see above)
 * B: some flags that indicate what kind of entry we have:
 *      0 - pure bullshit, this shouldn't happen
 *      1 - failed low (alpha-fail)
 *      2 - failed high (beta-fail)
 *      3 - provoked a new alpha -> exact value
 * C: 1 bit: value sign
 * D: 16 bits: currently evaluated value
 * E: 16 bits: depth of the search that gave us a value of C
 */
typedef struct {
	BITBOARD verification_key;
	BITBOARD data;
} HASHENTRY;


/*********************************************************************************
 * SITUATION: This main structure completely represents a situation on the board
 *   including all bitboards we need for the positional evaluation.
 *   In earlier versions, I had several instances of it - at the moment, only
 *   g->sit is used since I'm now able to store all the information about earlier
 *   positions in the hash table.
 */
typedef struct {
	signed char square[64];                    /* value between <PAWN> and <KING> */
	BITBOARD all_figures[3];                   /* * * * * * * * * * * * */
	BITBOARD pawns[2];                         /*  piece bitboards      */
	BITBOARD knights[2];                       /*  that are needed for  */
	BITBOARD bishops[2];                       /*  doing/undoing moves  */
	BITBOARD rooks[2];                         /*  and evaluating the   */
	BITBOARD queens[2];                        /*  current position.    */
	BITBOARD kings[2];                         /* * * * * * * * * * * * */
	BITBOARD zobrist;                          /* zobrist key of the current situation */
	int material_score[2];                     /* pure material score */
	int pawn_count[2];                         /* number of pawns of each side */
	int last_irreversible_move;                /* attention: counts half-moves */
	int rule50counter;                         /* attention: counts half-moves (<100) */
	char king[2];                              /* current king position */
	char color_to_move;                        /* either P_WHITE or P_BLACK */
	char enpassant;                            /* on which square may the enemy hit en passant? */
	/********************************************
	 * castling_possible (8 bits): |x|x|xxxxxx|
	 *                             |A|B|--C---|
	 * A: small castling possible
	 * B: great castling possible
	 */
	char castling_disallowed[2];
} SITUATION;


/*********************************************************************************
 * GAME: The biggest structure of all contains all the information needed in the
 *   current game. It includes search data, the current board situation <sit> and
 *   much more.
 *   Theoretically, (in a later version?) we might save a dump of its only instance
 *   <g> in a file to resume the search at a certain position.
 */ 
typedef struct {
	SITUATION sit;                             /* current board situation */
	SITBACKUP passed_backup[MAXPASSEDSITUATIONS];  /* used for '.u' in PHRI */
	BITBOARD passed_zobrist[MAXPASSEDSITUATIONS];  /* used for draw by repetition check */
	MOVE tempmoves[MAXPLY];                    /* only needed for debugging purposes */
	MOVE prec_move;                            /* preceding move (e.g. last user input) */
	MOVE best_root_move;                       /* best move at ply 0 */
	MOVE tlmove[MAXMOVES];                     /* all top level moves */
	MOVE pv[MAXPLY];                           /* current principal variation moves */
	MOVE first_killer[MAXPLY];                 /* killer heuristics: first move */
	MOVE second_killer[MAXPLY];                /* killer heuristics: second move */
	signed int phase;                          /* opening / middlegame / endgame? */
	long nodes;                                /* number of moves we've searched up to now */
	int tlworth[MAXMOVES];                     /* tentative worth of all top level moves */
	int tlcount;                               /* number of top level moves */
	int pv_length;                             /* length of the principal variation */
	int result;                                /* game result: see <const.h> for values */
	int ply;                                   /* current search ply */
	int seldepth;                              /* the maximum depth we've reached so far. */
	int fullmoves_number;                      /* pgn move number: e.g. "7. Sxf5" */
	int halfmove_index;                        /* half moves since the beginning */
	int last_destination[MAXPLY];              /* used to recapture the last moved piece */
	int root_castling_disallowed[2];           /* used for penalizing castling right loss */
} GAME;